import json
from hashlib import md5
from unittest.mock import patch
from uuid import UUID

from django.db.models import CharField, Value

from arkindex.documents.indexer import Indexer
from arkindex.documents.models import Corpus, EntityType, MetaType
from arkindex.process.models import WorkerRun
from arkindex.project.tests import FixtureTestCase


class TestIndexerCommand(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name="private", indexable=True)
        cls.worker_run = WorkerRun.objects.get(version__worker__slug="reco")
        cls.worker_version = cls.worker_run.version
        cls.worker = cls.worker_version.worker
        page_type, _ = cls.private_corpus.types.get_or_create(slug="page", display_name="Page", indexable=True)
        cls.page = cls.private_corpus.elements.create(name="New page", type=page_type, worker_run=cls.worker_run, worker_version=cls.worker_version)

    @patch("arkindex.documents.indexer.solr")
    def test_setup(self, mock_solr):
        mock_collections = mock_solr.collections
        mock_schema = mock_solr.schema
        mock_collections.exists.return_value = False
        mock_schema.does_field_exist.return_value = False

        indexer = Indexer("corpus_id")
        indexer.setup()

        self.assertEqual(mock_collections.exists.call_count, 1)
        (index_name, ), _ = mock_collections.exists.call_args
        self.assertEqual(index_name, indexer.collection_name)

        self.assertEqual(mock_collections.create.call_count, 1)
        (index_name, args), _ = mock_collections.create.call_args
        self.assertEqual(index_name, indexer.collection_name)
        self.assertEqual(args, indexer.solr_num_shards)

        self.assertEqual(mock_solr.transport.send_request.call_count, len(indexer.solr_type_fields))
        for type_field, (_, kwargs) in zip(indexer.solr_type_fields, mock_solr.transport.send_request.call_args_list):
            self.assertDictEqual(kwargs, {
                "method": "POST",
                "endpoint": mock_solr.schema.schema_endpoint,
                "collection": indexer.collection_name,
                "data": json.dumps({"add-field-type": type_field})
            })

        self.assertEqual(mock_schema.does_field_exist.call_count, len(indexer.solr_fields))
        self.assertEqual(mock_schema.create_field.call_count, len(indexer.solr_fields))
        for field, ((index_name, args), _) in zip(indexer.solr_fields, mock_schema.create_field.call_args_list):
            self.assertEqual(index_name, indexer.collection_name)
            self.assertDictEqual(args, field)
        self.assertEqual(mock_schema.replace_field.call_count, 0)

    @patch("arkindex.documents.indexer.solr")
    def test_already_setup(self, mock_solr):
        mock_collections = mock_solr.collections
        mock_schema = mock_solr.schema
        mock_collections.exists.return_value = True
        mock_schema.does_field_exist.return_value = True

        indexer = Indexer("corpus_id")
        indexer.setup()

        self.assertEqual(mock_collections.exists.call_count, 1)
        (index_name, ), _ = mock_collections.exists.call_args
        self.assertEqual(index_name, indexer.collection_name)

        self.assertEqual(mock_collections.create.call_count, 0)

        self.assertEqual(mock_solr.transport.send_request.call_count, len(indexer.solr_type_fields))
        for type_field, (_, kwargs) in zip(indexer.solr_type_fields, mock_solr.transport.send_request.call_args_list):
            self.assertDictEqual(kwargs, {
                "method": "POST",
                "endpoint": mock_solr.schema.schema_endpoint,
                "collection": indexer.collection_name,
                "data": json.dumps({"add-field-type": type_field})
            })

        self.assertEqual(mock_schema.does_field_exist.call_count, len(indexer.solr_fields))
        self.assertEqual(mock_schema.replace_field.call_count, len(indexer.solr_fields))
        self.assertEqual(mock_schema.create_field.call_count, 0)

    @patch("arkindex.documents.indexer.solr")
    def test_drop_index(self, mock_solr):
        indexer = Indexer("corpus_id")
        indexer.drop_index()

        self.assertEqual(mock_solr.delete_doc_by_query.call_count, 1)
        (index_name, query), kwargs = mock_solr.delete_doc_by_query.call_args
        self.assertEqual(index_name, indexer.collection_name)
        self.assertEqual(query, "*:*")
        self.assertDictEqual(kwargs, {"commit": True})

    def test_hash_worker(self):
        indexer = Indexer(None)
        self.assertIsNone(indexer.hash_worker(None))
        self.assertEqual(indexer.hash_worker(self.worker_run), self.worker.name)

    def test_build_id(self):
        expected = UUID(md5(self.page.id.bytes + self.page.id.bytes).hexdigest())
        indexer = Indexer(None)
        self.assertEqual(indexer.build_solr_id(self.page, self.page), expected)

    def test_build_element(self):
        annotated_pages = self.private_corpus.elements.filter(id=self.page.id).annotate(
            parent_id=Value(self.page.id, output_field=CharField()),
            parent_name=Value(self.page.name, output_field=CharField()),
            parent_type=Value(self.page.type.display_name, output_field=CharField()),
            type_name=Value(self.page.type.display_name, output_field=CharField())
        )
        indexer = Indexer(None)
        self.assertDictEqual(indexer.build_element(annotated_pages.first()), {
            "id": str(self.page.id),
            "parent_id": str(self.page.id),
            "parent_name": self.page.name,
            "parent_type": self.page.type.display_name,
            "element_id": str(self.page.id),
            "element_text": self.page.name,
            "element_type": self.page.type.display_name,
            "element_worker": self.worker.name,
            "element_image": None
        })

    def test_build_transcriptions(self):
        tr_1 = self.page.transcriptions.create(
            confidence=0.8,
            text="Transcription on the page",
            worker_run=self.worker_run,
            worker_version=self.worker_version
        )
        tr_2 = self.page.transcriptions.create(
            confidence=0.5,
            text="Second transcription",
            worker_run=self.worker_run,
            worker_version=self.worker_version
        )
        indexer = Indexer(None)
        self.assertListEqual(indexer.build_transcriptions(self.page, {"key": "value"}), [
            {
                "key": "value",
                "id": str(indexer.build_solr_id(self.page, tr_1)),
                "transcription_id": str(tr_1.id),
                "transcription_confidence": tr_1.confidence,
                "transcription_text": tr_1.text,
                "transcription_worker": self.worker.name,
            },
            {
                "key": "value",
                "id": str(indexer.build_solr_id(self.page, tr_2)),
                "transcription_id": str(tr_2.id),
                "transcription_confidence": tr_2.confidence,
                "transcription_text": tr_2.text,
                "transcription_worker": self.worker.name,
            }
        ])

    def test_build_classifications(self):
        cl_1 = self.page.classifications.create(
            confidence=0.8,
            worker_run=self.worker_run,
            worker_version=self.worker_version,
            ml_class=self.private_corpus.ml_classes.create(name="Cat")
        )
        cl_2 = self.page.classifications.create(
            confidence=0.4,
            worker_run=self.worker_run,
            worker_version=self.worker_version,
            ml_class=self.private_corpus.ml_classes.create(name="Dog")
        )
        indexer = Indexer(None)
        self.assertListEqual(indexer.build_classifications(self.page, {"key": "value"}), [
            {
                "key": "value",
                "id": str(indexer.build_solr_id(self.page, cl_1)),
                "classification_id": str(cl_1.id),
                "classification_name": cl_1.ml_class.name,
                "classification_text": cl_1.ml_class.name,
                "classification_confidence": cl_1.confidence,
                "classification_worker": self.worker.name,
            },
            {
                "key": "value",
                "id": str(indexer.build_solr_id(self.page, cl_2)),
                "classification_id": str(cl_2.id),
                "classification_name": cl_2.ml_class.name,
                "classification_text": cl_2.ml_class.name,
                "classification_confidence": cl_2.confidence,
                "classification_worker": self.worker.name,
            }
        ])

    def test_build_metadatas(self):
        self.private_corpus.allowed_metadatas.create(type=MetaType.Location, name="Country")
        self.private_corpus.allowed_metadatas.create(type=MetaType.Text, name="Folio")
        md_1 = self.page.metadatas.create(
            type=MetaType.Location,
            name="Country",
            value="France",
            worker_run=self.worker_run,
            worker_version=self.worker_version
        )
        md_2 = self.page.metadatas.create(
            type=MetaType.Text,
            name="Folio",
            value="1",
            worker_run=self.worker_run,
            worker_version=self.worker_version
        )
        indexer = Indexer(None)
        self.assertListEqual(indexer.build_metadatas(self.page, {"key": "value"}), [
            {
                "key": "value",
                "id": str(indexer.build_solr_id(self.page, md_1)),
                "metadata_id": str(md_1.id),
                "metadata_name": md_1.name,
                "metadata_text": md_1.value,
                "metadata_type": md_1.type.value,
                "metadata_worker": self.worker.name,
            },
            {
                "key": "value",
                "id": str(indexer.build_solr_id(self.page, md_2)),
                "metadata_id": str(md_2.id),
                "metadata_name": md_2.name,
                "metadata_text": md_2.value,
                "metadata_type": md_2.type.value,
                "metadata_worker": self.worker.name,
            }
        ])

    def test_build_entities(self):
        self.private_corpus.allowed_metadatas.create(type=MetaType.Location, name="Country")
        location_type = EntityType.objects.create(name="location", corpus=self.private_corpus)
        entity_1 = self.private_corpus.entities.create(name="CDLK", type=location_type, worker_run=self.worker_run,
            worker_version=self.worker_version)
        tr = self.page.transcriptions.create(
            confidence=0.8,
            text="Transcription on the page",
        )
        entity_1.transcription_entities.create(
            transcription=tr,
            offset=0,
            length=len(entity_1.name)
        )
        person_type = EntityType.objects.create(name="person", corpus=self.private_corpus)
        entity_2 = self.private_corpus.entities.create(name="Robert", type=person_type, worker_run=self.worker_run,
            worker_version=self.worker_version)
        self.page.metadatas.create(
            type=MetaType.Location,
            name="Country",
            value="France",
            entity=entity_2
        )
        indexer = Indexer(None)
        self.assertListEqual(indexer.build_entities(self.page, {"key": "value"}), [
            {
                "key": "value",
                "id": str(indexer.build_solr_id(self.page, entity_1)),
                "entity_id": str(entity_1.id),
                "entity_text": entity_1.name,
                "entity_type": "location",
                "entity_worker": self.worker.name,
            },
            {
                "key": "value",
                "id": str(indexer.build_solr_id(self.page, entity_2)),
                "entity_id": str(entity_2.id),
                "entity_text": entity_2.name,
                "entity_type": "person",
                "entity_worker": self.worker.name,
            }
        ])

    @patch("arkindex.documents.indexer.solr")
    def test_index(self, mock_solr):
        location_type = EntityType.objects.create(name="location", corpus=self.private_corpus)
        entity = self.private_corpus.entities.create(name="CDLK", type=location_type, worker_run=self.worker_run,
            worker_version=self.worker_version)
        tr = self.page.transcriptions.create(
            confidence=0.8,
            text="Transcription on the page",
        )
        entity.transcription_entities.create(
            transcription=tr,
            offset=0,
            length=len(entity.name)
        )

        indexer = Indexer(self.private_corpus.id)
        with self.assertExactQueries("indexer_prefetch.sql", params={
            "corpus_id": self.private_corpus.id.hex,
            "page_id": self.page.id.hex,
            "worker_run_id": self.worker_run.id.hex,
            "worker_version_id": self.worker_version.id.hex,
            "worker_id": self.worker.id.hex,
            "transcription_id": tr.id.hex,
            "type_id": location_type.id.hex,
        }):
            indexer.index()
        self.assertEqual(mock_solr.index.call_count, 1)
        (index_name, documents), kwargs = mock_solr.index.call_args
        self.assertEqual(index_name, indexer.collection_name)
        self.assertListEqual(documents, [
            {
                "id": str(self.page.id),
                "element_id": str(self.page.id),
                "element_text": self.page.name,
                "element_type": self.page.type.display_name,
                "element_worker": self.worker.name,
                "element_image": self.page.iiif_thumbnail_url,
                "parent_id": str(self.page.id),
                "parent_name": self.page.name,
                "parent_type": self.page.type.display_name,
            },
            {
                "id": str(indexer.build_solr_id(self.page, tr)),
                "element_id": str(self.page.id),
                "element_text": self.page.name,
                "element_type": self.page.type.display_name,
                "element_worker": self.worker.name,
                "element_image": self.page.iiif_thumbnail_url,
                "parent_id": str(self.page.id),
                "parent_name": self.page.name,
                "parent_type": self.page.type.display_name,
                "transcription_id": str(tr.id),
                "transcription_confidence": tr.confidence,
                "transcription_text": tr.text,
                "transcription_worker": None
            },
            {
                "id": str(indexer.build_solr_id(self.page, entity)),
                "element_id": str(self.page.id),
                "element_text": self.page.name,
                "element_type": self.page.type.display_name,
                "element_worker": self.worker.name,
                "element_image": self.page.iiif_thumbnail_url,
                "parent_id": str(self.page.id),
                "parent_name": self.page.name,
                "parent_type": self.page.type.display_name,
                "entity_id": str(entity.id),
                "entity_text": entity.name,
                "entity_type": "location",
                "entity_worker": self.worker.name,
            }]
        )
        self.assertDictEqual(kwargs, {"commit": True})
