from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import State
from arkindex.process.models import (
    ActivityState,
    Process,
    ProcessMode,
    WorkerActivity,
    WorkerActivityState,
    WorkerVersion,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestProcessDestroy(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = cls.corpus.processes.create(mode=ProcessMode.Workers, creator=cls.user)

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.delete(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_running(self):
        """
        It is not possible to delete a running import
        """
        self.client.force_login(self.user)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.process.tasks.update(state=State.Running)

        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"__all__": ["Cannot delete a process while it is running"]})

    def test_activities_pending(self):
        """
        It is not possible to delete an import with non initialized activities
        """
        self.client.force_login(self.user)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.process.activity_state = ActivityState.Pending
        self.process.save()

        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(
            response.json(),
            {"__all__": ["Please wait activities to be initialized before deleting this process"]}
        )

    @patch("arkindex.project.mixins.get_max_level", return_value=None)
    def test_process_no_permission(self, get_max_level_mock):
        """
        A user cannot delete a process linked to a corpus they have no access to
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertDictEqual(response.json(), {"detail": "Not found."})
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    @patch("arkindex.project.mixins.get_max_level", return_value=Role.Contributor.value)
    def test_process_not_admin(self, get_max_level_mock):
        """
        A user cannot delete a process linked to a corpus they have no admin access to
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have a sufficient access level to this process."})
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    @patch("arkindex.project.triggers.process_tasks.process_delete.delay")
    def test_worker_run_in_use(self, process_delay_mock):
        """
        A process with worker runs linked to data cannot be deleted
        """
        self.client.force_login(self.user)
        run = self.process.worker_runs.create(version=WorkerVersion.objects.first())

        page = self.corpus.elements.get(name="Volume 1, page 1r")
        metadata = page.metadatas.get()
        entity = self.corpus.entities.create(name="test", type=self.corpus.entity_types.create())
        transcription = page.transcriptions.first()
        transcription_entity = transcription.transcription_entities.create(entity=entity, offset=1, length=1)
        classification = page.classifications.create(ml_class=self.corpus.ml_classes.create(name="test"), confidence=1)

        for i, related_obj in enumerate((page, entity, transcription, transcription_entity, classification, metadata)):
            with self.subTest(related_obj=related_obj):
                related_obj.worker_run = run
                related_obj.worker_version = run.version
                related_obj.save()

                with self.assertNumQueries(5 + i):
                    response = self.client.delete(
                        reverse("api:process-details", kwargs={"pk": self.process.id})
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {
                    "__all__": ["Some worker runs on this process are linked to ML results and cannot be deleted."]
                })
                self.assertFalse(process_delay_mock.called)

            related_obj.worker_run, related_obj.worker_version = None, None
            related_obj.save()
            process_delay_mock.reset_mock()

    def test_corpus_import_admin(self):
        """
        A user is allowed to delete a process if he has an admin right to its corpus.
        If no activities exists for this process, it is deleted directly.
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(20):
            response = self.client.delete(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        with self.assertRaises(Process.DoesNotExist):
            self.process.refresh_from_db()

    def test_superuser(self):
        """
        A superuser is allowed to delete any process
        """
        self.client.force_login(self.superuser)

        with self.assertNumQueries(20):
            response = self.client.delete(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        with self.assertRaises(Process.DoesNotExist):
            self.process.refresh_from_db()

    @patch("arkindex.project.triggers.process_tasks.process_delete.delay")
    def test_corpus_import_async(self, delay_mock):
        """
        A process with activities is deleted in an async task and a HTTP_202_ACCEPTED is returned
        """
        self.client.force_login(self.user)
        self.process.activities.create(
            element=self.corpus.elements.get(name="Volume 1"),
            worker_version=WorkerVersion.objects.get(worker__slug="reco"),
        )

        with self.assertNumQueries(11):
            response = self.client.delete(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)

        self.assertEqual(delay_mock.call_count, 1)
        (process_arg,), kwargs = delay_mock.call_args
        self.assertEqual(process_arg.id, self.process.id)
        self.assertEqual(kwargs, {"delete_activity": WorkerActivityState.Queued})

    @patch("arkindex.project.triggers.process_tasks.process_delete.delay")
    def test_process_activities(self, delay_mock):
        """
        Deleting a process deletes the related activities with a queued state and
        sets the the FK on remaining activities to null
        Mock the async deletion task to run it directly in this test
        """
        from arkindex.process.tasks import process_delete
        delay_mock.side_effect = process_delete
        self.client.force_login(self.superuser)
        queued_activity = WorkerActivity.objects.create(
            element=self.corpus.elements.get(name="Volume 1"),
            process=self.process,
            worker_version=WorkerVersion.objects.get(worker__slug="reco"),
            state=WorkerActivityState.Queued,
        )
        completed_activity = WorkerActivity.objects.create(
            element=self.corpus.elements.get(name="Volume 1, page 2r"),
            process=self.process,
            worker_version=WorkerVersion.objects.get(worker__slug="reco"),
            state=WorkerActivityState.Processed,
        )

        with self.assertNumQueries(21):
            response = self.client.delete(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)

        with self.assertRaises(Process.DoesNotExist):
            self.process.refresh_from_db()
        with self.assertRaises(WorkerActivity.DoesNotExist):
            queued_activity.refresh_from_db()
        with self.assertRaises(WorkerActivity.DoesNotExist):
            completed_activity.refresh_from_db()
