from datetime import datetime, timezone
from unittest.mock import call, patch

from rest_framework import status
from rest_framework.reverse import reverse

from arkindex.documents.models import Corpus
from arkindex.process.models import (
    ArkindexFeature,
    FeatureUsage,
    ProcessMode,
    WorkerConfiguration,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersionState
from arkindex.users.models import Role


class TestApplyTemplate(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name="private")
        cls.private_corpus.memberships.create(user=cls.user, level=Role.Contributor.value)
        cls.private_corpus.elements.create(
            type=cls.private_corpus.types.create(slug="volume", folder=True),
            name="Hidden",
        )
        cls.private_template = cls.private_corpus.processes.create(
            creator=cls.user, mode=ProcessMode.Template
        )
        # Workers ProcessMode
        cls.version_1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_1 = cls.version_1.worker
        cls.version_2 = WorkerVersion.objects.get(worker__slug="dla")
        cls.worker_2 = cls.version_2.worker
        cls.version_3 = WorkerVersion.objects.get(worker__slug="worker-gpu")

        cls.process = cls.corpus.processes.create(
            creator=cls.user, mode=ProcessMode.Workers
        )
        cls.template = cls.corpus.processes.create(
            creator=cls.user, mode=ProcessMode.Template
        )
        cls.worker_configuration = WorkerConfiguration.objects.create(
            worker=cls.version_1.worker,
            name="A config",
            configuration={"param1": "value1"},
        )

        cls.model = Model.objects.create(name="moo")
        cls.model_version = cls.model.versions.create(state=ModelVersionState.Available)

        cls.template_run_1 = cls.template.worker_runs.create(
            version=cls.version_1,
            configuration=cls.worker_configuration,
        )
        cls.template_run_2 = cls.template.worker_runs.create(
            version=cls.version_2,
            parents=[cls.template_run_1.id],
            model_version=cls.model_version,
        )

        cls.private_template.worker_runs.create(
            version=cls.version_1,
            configuration=cls.worker_configuration,
        )

    @patch("arkindex.project.mixins.get_max_level", return_value=Role.Guest.value)
    def test_requires_contributor_rights_on_template(self, get_max_level_mock):
        """Raise 403 if the user does not have rights on template
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.private_template.id)}),
                data={"process_id": str(self.process.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(
            response.json(),
            {"detail": "You do not have a contributor access to this process."},
        )
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.private_corpus))

    @patch("arkindex.project.mixins.get_max_level", side_effect=[Role.Contributor.value, Role.Guest.value])
    def test_requires_contributor_rights_on_process(self, get_max_level_mock):
        """Raise 403 if the user does not have rights on the target process
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.private_template.id)}),
                data={"process_id": str(self.process.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(
            response.json(),
            {"detail": "You do not have a contributor access to this process."},
        )
        self.assertEqual(get_max_level_mock.call_args_list, [
            call(self.user, self.private_corpus),
            call(self.user, self.corpus),
        ])

    @patch("arkindex.project.mixins.get_max_level", return_value=Role.Contributor.value)
    @patch("arkindex.users.utils.get_max_level", return_value=Role.Guest.value)
    def test_requires_access_rights_all_workers(self, worker_max_level_mock, process_max_level_mock):
        """
        Raise 403 if the user does not have rights on all workers concerned
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.private_template.id)}),
                data={"process_id": str(self.process.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(
            response.json(),
            {"detail": "You do not have an execution access to this worker."},
        )
        self.assertEqual(process_max_level_mock.call_count, 1)
        self.assertEqual(process_max_level_mock.call_args, call(self.user, self.private_corpus))
        self.assertEqual(worker_max_level_mock.call_count, 1)
        self.assertEqual(worker_max_level_mock.call_args, call(self.user, self.worker_1))

    def test_already_applied(self):
        """Raise 400 if the process already has a template attached
        """
        process_with_template = self.corpus.processes.create(
            creator=self.user, mode=ProcessMode.Workers, template=self.template
        )
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
            data={"process_id": str(process_with_template.id)},
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"process_id": ["A template has already been applied to this process."]},
        )

    def test_without_a_template(self):
        """
        Raise 404 if the template is not a template
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:apply-process-template", kwargs={"pk": str(self.process.id)}),
            data={"process_id": str(self.process.id)},
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_apply(self):
        self.assertIsNotNone(self.version_2.docker_image_iid)
        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data={"process_id": str(self.process.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["template_id"], str(self.template.id))

        self.process.refresh_from_db()
        self.assertEqual(self.process.template, self.template)

        child_run, parent_run = self.process.worker_runs.select_related("version__worker").order_by("version__worker__slug")

        self.assertEqual(parent_run.process_id, self.process.id)
        self.assertEqual(parent_run.version_id, self.version_1.id)
        self.assertIsNone(parent_run.model_version_id)
        self.assertEqual(parent_run.configuration_id, self.worker_configuration.id)
        self.assertListEqual(parent_run.parents, [])

        self.assertEqual(child_run.process_id, self.process.id)
        self.assertEqual(child_run.version_id, self.version_2.id)
        self.assertEqual(child_run.model_version_id, self.model_version.id)
        self.assertIsNone(child_run.configuration_id)
        self.assertListEqual(child_run.parents, [parent_run.id])

    def test_excludes_init_elements(self):
        init_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        init_run = self.template.worker_runs.create(version=init_version)
        self.template_run_1.parents = [init_run.id]
        self.template_run_1.save()

        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data={"process_id": str(self.process.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["id"], str(self.process.id))
        self.assertEqual(data["template_id"], str(self.template.id))

        self.process.refresh_from_db()
        self.assertEqual(self.process.template, self.template)

        # Only two workers runs in the process
        self.assertEqual(self.template.worker_runs.count(), 3)
        self.assertEqual(self.process.worker_runs.count(), 2)

        child_run, parent_run = self.process.worker_runs.select_related("version__worker").order_by("version__worker__slug")

        self.assertEqual(parent_run.process_id, self.process.id)
        self.assertEqual(parent_run.version_id, self.version_1.id)
        self.assertIsNone(parent_run.model_version_id)
        self.assertEqual(parent_run.configuration_id, self.worker_configuration.id)
        self.assertListEqual(parent_run.parents, [])

        self.assertEqual(child_run.process_id, self.process.id)
        self.assertEqual(child_run.version_id, self.version_2.id)
        self.assertEqual(child_run.model_version_id, self.model_version.id)
        self.assertIsNone(child_run.configuration_id)
        self.assertListEqual(child_run.parents, [parent_run.id])

        # No elements initialisation run in the created process
        self.assertFalse(self.process.worker_runs.filter(version=init_version).exists())

    def test_delete_previous_worker_runs(self):
        self.client.force_login(self.user)
        # Create a process with one worker run already
        process = self.corpus.processes.create(
            creator=self.user, mode=ProcessMode.Workers
        )
        process.worker_runs.create(version=self.version_2)
        # Apply a template that has two other worker runs
        with self.assertNumQueries(12):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data={"process_id": str(process.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["template_id"], str(self.template.id))

        process.refresh_from_db()
        self.assertEqual(process.template, self.template)

        # Assert that the previous worker runs was deleted and the template was correctly applied
        child_run, parent_run = process.worker_runs.select_related("version__worker").order_by("version__worker__slug")

        self.assertEqual(parent_run.process_id, process.id)
        self.assertEqual(parent_run.version_id, self.version_1.id)
        self.assertIsNone(parent_run.model_version_id)
        self.assertEqual(parent_run.configuration_id, self.worker_configuration.id)
        self.assertListEqual(parent_run.parents, [])

        self.assertEqual(child_run.process_id, process.id)
        self.assertEqual(child_run.version_id, self.version_2.id)
        self.assertEqual(child_run.model_version_id, self.model_version.id)
        self.assertIsNone(child_run.configuration_id)
        self.assertListEqual(child_run.parents, [parent_run.id])

    def test_use_gpu(self):
        self.client.force_login(self.user)

        # Set invalid values: the version with disabled GPU usage gets a GPU
        self.template.worker_runs.filter(version=self.version_1).update(use_gpu=True)
        # A signal is trying to set use_gpu to the correct values, so we create then update to give no GPU to a version that requires a GPU
        self.template.worker_runs.create(version=self.version_3)
        self.template.worker_runs.filter(version=self.version_3).update(use_gpu=False)

        # Have two runs with a version that supports GPU usage, to test that both True and False are copied
        self.version_2.gpu_usage = FeatureUsage.Supported
        self.version_2.save()
        self.template.worker_runs.create(version=self.version_2, configuration=self.worker_configuration, use_gpu=True)

        self.assertQuerySetEqual((
            self.template.worker_runs
            .order_by("version__gpu_usage", "use_gpu")
            .values_list("version__gpu_usage", "use_gpu")
        ), [
            (FeatureUsage.Disabled, True),
            (FeatureUsage.Required, False),
            (FeatureUsage.Supported, False),
            (FeatureUsage.Supported, True),
        ])

        with self.assertNumQueries(10):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                {"process_id": str(self.process.id)},
            )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertQuerySetEqual((
            self.process.worker_runs
            .order_by("version__gpu_usage", "use_gpu")
            .values_list("version__gpu_usage", "use_gpu")
        ), [
            (FeatureUsage.Disabled, False),
            (FeatureUsage.Required, True),
            (FeatureUsage.Supported, False),
            (FeatureUsage.Supported, True),
        ])


    def test_unavailable_worker_version(self):
        self.version_2.state = WorkerVersionState.Error
        self.version_2.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data={"process_id": str(self.process.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["This template contains one or more unavailable worker versions and cannot be applied."]})
        self.process.refresh_from_db()
        self.assertEqual(self.process.template, None)

    def test_archived_worker(self):
        self.worker_2.archived = datetime.now(timezone.utc)
        self.worker_2.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data={"process_id": str(self.process.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["This template contains one or more worker versions from archived workers and cannot be applied."]})
        self.process.refresh_from_db()
        self.assertEqual(self.process.template, None)

    def test_unavailable_model_version(self):
        self.model_version.state = ModelVersionState.Error
        self.model_version.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data={"process_id": str(self.process.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"non_field_errors": ["This template contains one or more unavailable model versions and cannot be applied."]})

        self.process.refresh_from_db()
        self.assertEqual(self.process.template, None)

    def test_archived_model(self):
        self.model.archived = datetime.now(timezone.utc)
        self.model.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data={"process_id": str(self.process.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"non_field_errors": ["This template contains one or more model versions from archived models and cannot be applied."]})
        self.process.refresh_from_db()
        self.assertEqual(self.process.template, None)

    def test_unsupported_mode(self):
        self.client.force_login(self.user)
        for mode in set(ProcessMode) - {ProcessMode.Workers, ProcessMode.Dataset, ProcessMode.Local}:
            with self.subTest(mode=mode):
                self.process.mode = mode
                self.process.save()

                with self.assertNumQueries(5):
                    response = self.client.post(
                        reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                        data={"process_id": str(self.process.id)},
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "process_id": ["Templates can only be applied on Workers or Dataset processes."],
                })
                self.process.refresh_from_db()
                self.assertEqual(self.process.template, None)

    def test_local(self):
        self.client.force_login(self.user)
        local_process = self.user.processes.get(mode=ProcessMode.Local)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:apply-process-template", kwargs={"pk": str(self.template.id)}),
                data={"process_id": str(local_process.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "process_id": ["Process with this ID does not exist."],
        })
        local_process.refresh_from_db()
        self.assertEqual(local_process.template, None)
