from unittest.mock import call, patch
from uuid import uuid4

from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import ACTIVE_STATES, State
from arkindex.process.models import ActivityState, ProcessMode, WorkerActivity, WorkerActivityState, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase


class TestSelectProcessFailures(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = cls.corpus.processes.create(mode=ProcessMode.Workers, creator=cls.user)
        cls.recognizer = WorkerVersion.objects.get(worker__slug="reco")

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse("api:process-select-failures", kwargs={"pk": str(self.process.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:process-select-failures", kwargs={"pk": str(self.process.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_valid_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:process-select-failures", kwargs={"pk": str(uuid4())}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch("arkindex.project.mixins.get_max_level", return_value=None)
    def test_requires_corpus_read_access(self, get_max_level_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:process-select-failures", kwargs={"pk": str(self.process.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have a read access to the corpus of this process."})
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    def test_requires_workers_mode(self):
        self.client.force_login(self.user)
        self.user.processes.filter(mode=ProcessMode.Local).delete()

        for mode in set(ProcessMode) - {ProcessMode.Workers}:
            with self.subTest(mode=mode):
                self.process.mode = mode
                self.process.corpus = None if mode == ProcessMode.Local else self.corpus
                self.process.save()

                with self.assertNumQueries(3):
                    response = self.client.post(reverse("api:process-select-failures", kwargs={"pk": str(self.process.id)}))
                    self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_requires_worker_activity(self):
        self.client.force_login(self.user)
        self.process.activity_state = ActivityState.Disabled
        self.process.save()

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:process-select-failures", kwargs={"pk": str(self.process.id)}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(
            response.json(),
            {"__all__": ["The process must use worker activities to select elements with failures."]},
        )

    def test_requires_finished_state(self):
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.assertEqual(self.process.tasks.get().run, 0)
        self.process.activity_state = ActivityState.Ready
        self.process.save()
        self.client.force_login(self.user)

        for state in ACTIVE_STATES:
            with self.subTest(state=state):
                self.process.tasks.update(state=state)
                self.assertEqual(self.process.state, state)

                with self.assertNumQueries(3):
                    response = self.client.post(
                        reverse("api:process-select-failures", kwargs={"pk": str(self.process.id)})
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(
                    response.json(),
                    {"__all__": ["The process must be finished to select elements with failures."]},
                )

    def test_no_failure(self):
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.assertEqual(self.process.tasks.get().run, 0)
        self.process.tasks.update(state=State.Completed)
        self.process.activity_state = ActivityState.Ready
        self.process.save()
        # Process has a single element that is ok
        elt = self.corpus.elements.get(name="Volume 1")
        self.process.elements.add(elt.id)
        WorkerActivity.objects.create(
            process=self.process,
            worker_version=self.recognizer,
            element=elt,
            state=WorkerActivityState.Processed,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:process-select-failures", kwargs={"pk": str(self.process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"__all__": ["This process has no element in a failure state."]},
        )

    def test_select(self):
        # Failures in a previous run should not prevent calling this endpoint
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.process.tasks.update(state=State.Error)
        # Reload the prefetched worker runs because a new init_elements run was created
        # and running again would try to recreate it
        self.process.refresh_from_db()
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.process.tasks.filter(run=1).update(state=State.Completed)

        self.process.activity_state = ActivityState.Ready
        self.process.save()

        elements = self.corpus.elements.filter(
            name__in=(
                "Volume 1, page 1r",
                "Volume 1, page 1v",
                "Volume 1, page 2r",
            )
        )
        # Only recto pages has an activity with a failure
        for elt in elements:
            self.process.elements.add(elt.id)
            WorkerActivity.objects.create(
                process=self.process,
                worker_version=self.recognizer,
                element=elt,
                state=WorkerActivityState.Processed if elt.name.endswith("v") else WorkerActivityState.Error,
            )

        self.assertCountEqual(list(self.user.selections.values_list("element__name", flat=True)), [])
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:process-select-failures", kwargs={"pk": str(self.process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertCountEqual(
            list(self.user.selections.values_list("element__name", flat=True)),
            ["Volume 1, page 1r", "Volume 1, page 2r"],
        )
