from datetime import datetime, timezone
from enum import Enum
from urllib.parse import quote, unquote

import bleach
from django.contrib.gis.geos import Point
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers

from arkindex.documents.models import MetaType
from arkindex.ponos.utils import get_process_from_task_auth
from arkindex.process.models import ProcessMode, WorkerRun
from arkindex.project.gis import ensure_linear_ring
from arkindex.project.tools import is_prefetched


class EnumField(serializers.ChoiceField):
    """
    Serializes an enum field into a JSON value
    """

    def __init__(self, enum, *args, **kwargs):
        assert issubclass(enum, Enum) or not enum
        self.enum = enum
        choices = [(item.value, item.name) for item in self.enum]
        super().__init__(choices, *args, **kwargs)

    def to_representation(self, obj):
        if not isinstance(obj, self.enum):
            obj = self.to_internal_value(obj)
        return obj.value

    def to_internal_value(self, data):
        assert self.enum is not None, "No enum set on EnumField"
        try:
            return self.enum(data)
        except ValueError:
            raise serializers.ValidationError(f"Value is not of type {self.enum.__name__}")


class PointField(serializers.ListField):
    child = serializers.IntegerField(min_value=0)

    def __init__(self, *args, **kwargs):
        kwargs.update(min_length=2, max_length=2)
        super().__init__(*args, **kwargs)

    def run_validation(self, data):
        try:
            return Point(*super().run_validation(data))
        except (AssertionError, TypeError, ValueError) as e:
            raise serializers.ValidationError(str(e))


class LinearRingField(serializers.ListField):
    child = PointField()

    def __init__(self, *args, **kwargs):
        kwargs.update(min_length=3)
        super().__init__(*args, **kwargs)

    def run_validation(self, data):
        value = super().run_validation(data)

        # Ignore nothingness
        if value is None:
            return

        # Ensure the LineString is closed
        if value[0] != value[-1]:
            value.append(value[0])

        try:
            return ensure_linear_ring(value)
        except (AssertionError, ValueError) as e:
            raise serializers.ValidationError(str(e))


class IIIFPathField(serializers.CharField):
    """
    An IIIF image path field where special characters are escaped,
    including slashes.
    """

    def to_internal_value(self, data):
        value = super().to_internal_value(data)
        # If unquoting the value gives a different value, the value was already quoted
        if "/" in value or unquote(value) == value:
            value = quote(value, safe="")
        return value


METADATA_ALLOWED_HTML_TAGS = [
    "a", "b", "blockquote", "body", "br", "div",
    "em", "h1", "h2", "h3", "html", "i", "iframe",
    "img", "li", "marquee", "ol", "p", "sup",
    "table", "tbody", "td", "th", "thead", "tr",
    "ul"
]


@extend_schema_field({
    "oneOf": [
        {"type": "string"},
        {"type": "number"},
    ]
})
class MetaDataValueField(serializers.CharField):

    def get_attribute(self, instance):
        """
        Make the whole MetaData available to to_representation
        so that it can do MetaType-specific serialization
        """
        return instance

    def to_representation(self, instance):
        if instance.type == MetaType.Numeric:
            return float(instance.value)

        # Clean the metadata value of HTML tags, only for markdown metadata
        if instance.type == MetaType.Markdown:
            return bleach.clean(instance.value, tags=METADATA_ALLOWED_HTML_TAGS)

        return instance.value


# Hide this field from the OpenAPI schema
@extend_schema_field(None)
class ForbiddenField(serializers.Field):
    """
    A field that always causes a validation error.
    """

    def __init__(self, *fields, message="This field is forbidden."):
        self.message = message
        super().__init__(write_only=True, required=False, default=None)

    def run_validation(self, data=serializers.empty):
        # This method might be called even when this field is not set.
        if data is not serializers.empty:
            raise serializers.ValidationError(self.message)
        return super().run_validation(data)


class WorkerRunIDField(serializers.PrimaryKeyRelatedField):

    def __init__(self, **kwargs):
        assert "queryset" not in kwargs, "The `queryset` argument is automatically set by this field."
        assert "style" not in kwargs, "The `style` argument is automatically set by this field."

        # Allow overriding the source just in case, but default to `worker_run`, as we may need a different model field somewhere else
        if not kwargs.get("source"):
            kwargs["source"] = "worker_run"

        super().__init__(
            **kwargs,
            queryset=(
                WorkerRun
                .objects
                .select_related("process")
                # We do not need to load every single attribute from the run and process:
                # just get those we need to validate, and the summary that the serializers can return
                .only(
                    "id",
                    "version_id",
                    "process_id",
                    "process__mode",
                    "process__creator_id",
                    "summary",
                )
            ),
            # Make sure we do not fetch and display every WorkerRun in the database when opening an endpoint in the browser
            style={"base_template": "input.html"},
        )

    def bind(self, field_name, parent):
        assert field_name == "worker_run_id", "To maintain API consistency, the field name of WorkerRun IDs should always be `worker_run_id`."
        super().bind(field_name, parent)

    def to_internal_value(self, data):
        assert self.parent, "A parent serializer is required to validate a WorkerRun."
        request = self.parent.context.get("request")
        assert request, "An HTTP request is required in the serializer context to validate a WorkerRun."

        worker_run = super().to_internal_value(data)

        # Ignore nothingness
        if worker_run is None or worker_run is serializers.empty:
            return None

        # No matter the authentication, a user can use WorkerRuns of their own local process
        if worker_run.process.mode == ProcessMode.Local and worker_run.process.creator_id == request.user.id:
            return worker_run

        # In other cases, Ponos authentication is required,
        # and only the WorkerRuns of the process of the authenticated task are allowed.
        process = get_process_from_task_auth(request)
        if process is None:
            raise serializers.ValidationError("Ponos task authentication is required to use a WorkerRun of a process other than the user's local process.")

        if process.id != worker_run.process_id:
            raise serializers.ValidationError("Only the WorkerRuns of the authenticated task's process may be used.")

        return worker_run


class ArchivedField(serializers.BooleanField):
    """
    Exposes a boolean value to users, but stores it as either the current date when true, or None when false.

    The model class must have a `.is_archivable(user)` method,
    which returns True when the user is allowed to archive or unarchive this instance.

    This allows storing an archival date without letting API users control it,
    so that instances can be deleted automatically after a delay.
    """

    def __init__(self, *args, **kwargs):
        # We only handle true and false!
        assert not kwargs.get("allow_null"), "`allow_null` is not supported on ArchivedField."
        super().__init__(*args, **kwargs)

    def run_validation(self, data=serializers.empty):
        # Validates the boolean value and returns a datetime or None
        data = super().run_validation(data)

        if data is serializers.empty:
            return data

        request = self.context.get("request")
        assert request, "An HTTP request is required in the serializer context to validate an ArchivedField."

        instance = self.parent.instance
        # Creating something as archived is not supported
        if data is not None and instance is None:
            raise serializers.ValidationError("Only already existing instances can be marked as archived.")

        # Allow updating to the current value without any permissions, but if the value changes, require `is_archivable`.
        # Since the above call to super().run_validation() returns the current datetime, even if we didn't change
        # whether or not the worker is archived, the dates would be different. So we just compare whether or not
        # the instance's value and the new value are none.
        if self.get_attribute(instance) != (data is not None) and not instance.is_archivable(self.context["request"].user):
            raise serializers.ValidationError(f"You are not allowed to archive or unarchive this {instance._meta.verbose_name}.")

        return data

    def to_internal_value(self, data):
        value = super().to_internal_value(data)
        if value is True:
            return datetime.now(timezone.utc)
        if value is False:
            return None
        raise NotImplementedError

    def get_attribute(self, instance) -> bool:
        """
        Serializer.to_representation calls Field.get_attribute to retrieve
        the value of the attribute on an existing instance.
        If it is None, it does not call Field.to_representation,
        so that fields do not have to handle that.

        But we do want to handle it here, so we do within get_attribute itself.
        """
        return super().get_attribute(instance) is not None

    def to_representation(self, value) -> bool:
        if isinstance(value, bool):
            return value
        return value is not None


class DatasetSetsCountField(serializers.DictField):
    """
    Serialize the number of element per set on a dataset.

    This field is None, unless the sets have been prefetched
    with a `element_count` annotation holding the number of elements per set.
    """

    def __init__(self, **kwargs):
        super().__init__(
            child=serializers.IntegerField(),
            read_only=True,
            allow_null=True,
            **kwargs,
        )

    def get_attribute(self, instance):
        # Skip this field if sets are not prefetched, or if they are missing a count
        if (
            not is_prefetched(instance.sets)
            or not all(hasattr(set, "element_count") for set in instance.sets.all())
        ):
            return None

        return {
            set.name: set.element_count
            for set in instance.sets.all()
        }


class NullField(serializers.CharField):

    def to_representation(self, value):
        return None

    def to_internal_value(self, data):
        if data is not None:
            self.fail("invalid")
        return
